@@ -2,16 +2,17 @@
 import re
 import json
 import random
-from datetime import datetime
+from datetime import datetime, timedelta
 from json import JSONDecodeError
 
 import requests
 
-__version__ = "0.5.0"
+__version__ = "0.5.2"
 __author__ = "Kartik Talwar"
 __email__ = "hi@kartikt.com"
 __url__ = "https://github.com/KartikTalwar/duolingo"
 
+
 class Struct:
     def __init__(self, **entries):
         self.__dict__.update(entries)
@@ -33,8 +34,17 @@ class CaptchaException(DuolingoException):
     pass
 
 
+class OtherUserException(DuolingoException):
+    """
+    This exception is raised when set_username() has been called to get info on another user, but a method has then
+    been used which cannot give data on that new user.
+    """
+    pass
+
+
 class Duolingo(object):
-    USER_AGENT = "Duolingo API/{}".format(__version__)
+    USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.116 " \
+                 "Safari/537.36"
 
     def __init__(self, username, password=None, *, jwt=None, session_file=None):
         """
@@ -45,6 +55,7 @@ class Duolingo(object):
         requests.
         """
         self.username = username
+        self._original_username = username
         self.password = password
         self.session_file = session_file
         self.session = requests.Session()
@@ -118,6 +129,15 @@ class Duolingo(object):
         resp = self._make_req(self.get_user_url())
         return resp.status_code == 200
 
+    def get_user_url_by_id(self, fields=None):
+        if fields is None:
+            fields = []
+        url = 'https://www.duolingo.com/2017-06-30/users/{}'.format(self.user_data.id)
+        fields_params = requests.utils.requote_uri(','.join(fields))
+        if fields_params:
+            url += '?fields={}'.format(fields_params)
+        return url
+
     def get_user_url(self):
         return "https://duolingo.com/users/%s" % self.username
 
@@ -221,6 +241,18 @@ class Duolingo(object):
         except ValueError:
             raise DuolingoException('Failed to switch language')
 
+    def get_data_by_user_id(self, fields=None):
+        """
+        Get user's data from ``https://www.duolingo.com/2017-06-30/users/<user_id>``.
+        """
+        if fields is None:
+            fields = []
+        get = self._make_req(self.get_user_url_by_id(fields))
+        if get.status_code == 404:
+            raise DuolingoException('User not found')
+        else:
+            return get.json()
+
     def _get_data(self):
         """
         Get user's data from ``https://www.duolingo.com/users/<username>``.
@@ -459,6 +491,9 @@ class Duolingo(object):
 
     def get_vocabulary(self, language_abbr=None):
         """Get overview of user's vocabulary in a language."""
+        if self.username != self._original_username:
+            raise OtherUserException("Vocab cannot be listed when the user has been switched.")
+
         if language_abbr and not self._is_current_language(language_abbr):
             self._switch_language(language_abbr)
 
@@ -565,13 +600,20 @@ class Duolingo(object):
                 continue
             resp_data = resp.json()
             for challenge in resp_data['challenges']:
-                self._add_to_voice_url_dict(lang_abbr, challenge['prompt'], challenge['tts'])
+                if "prompt" in challenge and "tts" in challenge:
+                    self._add_to_voice_url_dict(lang_abbr, challenge['prompt'], challenge['tts'])
                 if challenge.get("metadata") and challenge['metadata'].get("non_character_tts"):
                     for word, url in challenge['metadata']['non_character_tts']['tokens'].items():
                         self._add_to_voice_url_dict(lang_abbr, word, url)
-                for token in challenge['tokens']:
-                    if token.get("tts") and token.get("value"):
-                        self._add_to_voice_url_dict(lang_abbr, token['value'], token['tts'])
+                if "tokens" in challenge:
+                    self._add_token_list_to_voice_url_dict(lang_abbr, challenge["tokens"])
+
+    def _add_token_list_to_voice_url_dict(self, lang_abbr, token_list):
+        for token in token_list:
+            if isinstance(token, list):
+                self._add_token_list_to_voice_url_dict(lang_abbr, token)
+            if isinstance(token, dict) and token.get("tts") and token.get("value"):
+                self._add_to_voice_url_dict(lang_abbr, token['value'], token['tts'])
 
     def _add_to_voice_url_dict(self, lang_abbr, word, url):
         word = word.lower()
@@ -608,11 +650,7 @@ class Duolingo(object):
             raise Exception('Could not get word definition')
 
     def get_daily_xp_progress(self):
-        daily_progress_url = \
-            "https://www.duolingo.com/2017-06-30/users/" \
-            "{}?fields=xpGoal,xpGains,streakData".format(self.user_data.id)
-        daily_progress_request = self._make_req(daily_progress_url)
-        daily_progress = daily_progress_request.json()
+        daily_progress = self.get_data_by_user_id(["xpGoal", "xpGains", "streakData"])
 
         if not daily_progress:
             raise DuolingoException(
@@ -621,9 +659,17 @@ class Duolingo(object):
 
         # xpGains lists the lessons completed on the last day where lessons were done.
         # We use the streakData.updatedTimestamp to get the last "midnight", and get lessons after that.
-        last_midnight = daily_progress['streakData']['updatedTimestamp']
+        reported_timestamp = daily_progress['streakData']['updatedTimestamp']
+        reported_midnight = datetime.fromtimestamp(reported_timestamp)
+        midnight = datetime.fromordinal(datetime.today().date().toordinal())
 
-        lessons = [lesson for lesson in daily_progress['xpGains'] if lesson['time'] > last_midnight]
+        # Sometimes the update is marked into the future. When this is the case
+        # we fall back on the system time for midnight.
+        time_discrepancy = min(midnight - reported_midnight, timedelta(0))
+        update_cutoff = round((reported_midnight + time_discrepancy).timestamp())
+
+        lessons = [lesson for lesson in daily_progress['xpGains'] if
+                lesson['time'] > update_cutoff]
 
         return {
             "xp_goal": daily_progress['xpGoal'],
